/*
Module : Finger.cpp
Purpose: Implementation for a MFC class encapsulation of the Finger protocol
Created: PJN / 11-10-1999
History: PJN / 17-10-1999 1. Preserved GetLastError state in CPJNFingerSocket::Close
         PJN / 03-11-2003 1. Updated copyright messages
                          2. Simplified the code in CPJNFinger::ReadResponse. Thanks to 
                          Clarke Brunt for reporting this issue.
         PJN / 14-04-2006 1. Updated copyright details.
                          2. Addition of a CPJNFINGER_EXT_CLASS macro which makes the class
                          easier to use in an extension DLL.
                          3. Renamed the class to CPJNFinger.
                          4. Module has been renamed to PJNFinger
                          5. Now uses the author's CWSocket sockets wrapper class. This now allows
                          the code to support connecting via Socks4, Socks5 and HTTP proxies.
                          6. Updated documentation to use the same style as my web site.
                          7. Did a spell check of the documentation.
                          8. Fixed a graceful disconnect problem in ReadResponse method.
                          9. Fixed a issue in CPJNFinger::Finger when code is compiled with /Wp64
         PJN / 28-12-2006 1. Updated the code to compile cleanly on VC 2005
                          2. Code now uses newer C++ style casts instead of C style casts.
                          3. Provided a GetLastError method which returns the last error which 
                          the code encountered. This is required because due to C++ destructors,
                          the value stored in GetLastError can be destroyed in the CPJNFinger
                          methods.

Copyright (c) 1999 - 2006 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/

//////////////// Includes ////////////////////////////////////////////

#include "stdafx.h"
#include "PJNFinger.h"


//////////////// Macros / Locals /////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


//////////////// Implementation //////////////////////////////////////

CPJNFinger::CPJNFinger() : m_dwTimeout(5000), 
                           m_ProxyType(ptNone), 
                           m_nProxyPort(1080),
                           m_dwLastError(ERROR_SUCCESS)
{
}

BOOL CPJNFinger::Finger(LPCTSTR pszHostName, LPCTSTR pszUser, CString& sReply, BOOL bVerbose, int nPort)
{
	USES_CONVERSION;

  //Create the socket
  CWSocket socket;
  try
  {
    socket.Create();
  }
  catch(CWSocketException* pEx)
  {
    TRACE(_T("CPJNFinger::Finger, Failed to create client socket, Error:%d\n"), pEx->m_nError);

    //Tidy up prior to returning
    m_dwLastError = pEx->m_nError;
    pEx->Delete();

    return FALSE;
  }

  //Connect to the finger Host
  try
  {
    //Bind if required
    if (m_sLocalBoundAddress.GetLength())
      socket.Bind(0, m_sLocalBoundAddress);

    switch (m_ProxyType)
    {
      case ptSocks4:
      {
        socket.ConnectViaSocks4(pszHostName, nPort, m_sProxyServer, m_nProxyPort, m_dwTimeout);
        break;
      }
      case ptSocks5:
      {
        if (m_sProxyUserName.GetLength())
          socket.ConnectViaSocks5(pszHostName, nPort, m_sProxyServer, m_nProxyPort, m_sProxyUserName, m_sProxyPassword, m_dwTimeout, FALSE);
        else
          socket.ConnectViaSocks5(pszHostName, nPort, m_sProxyServer, m_nProxyPort, NULL, NULL, m_dwTimeout, FALSE);
        break;
      }
      case ptHTTP:
      {
        CString sProxyResponse;
        if (m_sProxyUserName.GetLength())
        {
          if (m_sUserAgent.GetLength())
            socket.ConnectViaHTTPProxy(pszHostName, nPort, m_sProxyServer, m_nProxyPort, sProxyResponse, m_sProxyUserName, m_sProxyPassword, m_dwTimeout, m_sUserAgent);
          else
            socket.ConnectViaHTTPProxy(pszHostName, nPort, m_sProxyServer, m_nProxyPort, sProxyResponse, m_sProxyUserName, m_sProxyPassword, m_dwTimeout, NULL);
        }
        else
        {
          if (m_sUserAgent.GetLength())
            socket.ConnectViaHTTPProxy(pszHostName, nPort, m_sProxyServer, m_nProxyPort, sProxyResponse, NULL, NULL, m_dwTimeout, m_sUserAgent);
          else
            socket.ConnectViaHTTPProxy(pszHostName, nPort, m_sProxyServer, m_nProxyPort, sProxyResponse, NULL, NULL, m_dwTimeout, NULL);
        }
        break;
      }
      case ptNone:
      {
        socket.Connect(pszHostName, nPort);
        break;
      }
      default:
      {
        ASSERT(FALSE);
        break;
      }
    }
  }
  catch(CWSocketException* pEx)
  {
    TRACE(_T("CPJNFinger::Finger, Could not connect to the finger server, Error:%d\n"), pEx->m_nError);

    //Tidy up prior to returning
    m_dwLastError = pEx->m_nError;
    pEx->Delete();

    return FALSE;
  }

  //Generate the appropiate request
  CString sRequest;
  if (bVerbose)
  {
    if (pszUser)
      sRequest.Format(_T("/W %s\r\n"), pszUser);
    else
      sRequest = _T("/W\r\n");
  }
  else
  {
    if (pszUser)
      sRequest.Format(_T("%s\r\n"), pszUser);
    else
      sRequest = _T("\r\n");
  }

  //Send the request to the finger server
  LPSTR pszRequest = T2A(const_cast<LPTSTR>(sRequest.operator LPCTSTR()));
  int nCmdLength = static_cast<int>(strlen(pszRequest));
  try
  {
    socket.Send(pszRequest, nCmdLength);
  }
  catch(CWSocketException* pEx)
  {
    TRACE(_T("CPJNFinger::Finger, Failed to send request to the finger server, Error:%d\n"), pEx->m_nError);

    //Tidy up prior to returning
    m_dwLastError = pEx->m_nError;
    pEx->Delete();

    return FALSE;
  }

  //Read the response back from the host
  LPSTR pszOverFlowBuffer = NULL;
  char sBuf[4096];
  BOOL bSuccess = ReadResponse(socket, sBuf, 4096, &pszOverFlowBuffer, m_dwTimeout);
  if (pszOverFlowBuffer)
  {
    sReply = pszOverFlowBuffer;
    delete [] pszOverFlowBuffer;
  }
  else
    sReply = sBuf;

  if (bSuccess)
    m_dwLastError = ERROR_SUCCESS;

  return bSuccess;
}

BOOL CPJNFinger::ReadResponse(CWSocket& socket, LPSTR pszBuffer, int nInitialBufSize, LPSTR* ppszOverFlowBuffer, DWORD dwTimeout, int nGrowBy)
{
  ASSERT(ppszOverFlowBuffer);          //Must have a valid string pointer
  ASSERT(*ppszOverFlowBuffer == NULL); //Initially it must point to a NULL string

  //The local variables which will receive the data
  LPSTR pszRecvBuffer = pszBuffer;
  int nBufSize = nInitialBufSize;
  
  //retrieve the reponse using until we
	//get the terminator or a timeout occurs
	BOOL bFinished = FALSE;
	int nReceived = 0;
	while (!bFinished)
	{
    //check for readability
    BOOL bReadible = FALSE;
    try
    {
      bReadible = socket.IsReadible(dwTimeout);
    }
    catch(CWSocketException* pEx)
    {
	    pszRecvBuffer[nReceived] = '\0';

      m_dwLastError = pEx->m_nError;
      pEx->Delete();

			return FALSE;
    }
    
    if (!bReadible) //no data to receive, finish the loop
    {
		  pszRecvBuffer[nReceived] = '\0';
      bFinished = TRUE;
      continue;
    }

		//receive the data from the socket
    int nBufRemaining = nBufSize - nReceived - 1; //Allows allow one space for the NULL terminator
    if (nBufRemaining < 0)
      nBufRemaining = 0;

	  int nData = 0;
    try
    {
      nData = socket.Receive(pszRecvBuffer + nReceived, nBufRemaining);
    }
    catch(CWSocketException* pEx)
    {
      //Null terminate the data received
      if (pszRecvBuffer)
        pszRecvBuffer[nReceived] = '\0';

      m_dwLastError = pEx->m_nError;
      pEx->Delete();

      return FALSE;
    }

    if (nData < 1)
    {
      //NULL terminate the data received
      if (pszRecvBuffer)
		    pszBuffer[nReceived] = '\0';

		  return FALSE; 
    }
    else
    {
      //Increment the count of data received
		  nReceived += nData;							   

      //NULL terminate the data received
      if (pszRecvBuffer)
		    pszRecvBuffer[nReceived] = '\0';

      if ((nBufRemaining-nData) == 0) //No space left in the current buffer
      {
        //Allocate the new receive buffer
        nBufSize += nGrowBy; //Grow the buffer by the specified amount
        LPSTR pszNewBuf = new char[nBufSize];

        //copy the old contents over to the new buffer and assign 
        //the new buffer to the local variable used for retreiving 
        //from the socket
        if (pszRecvBuffer)
        {
          memcpy(pszNewBuf, pszRecvBuffer, nReceived);
          pszNewBuf[nReceived] = '\0';
        }
        pszRecvBuffer = pszNewBuf;

        //delete the old buffer if it was allocated
        if (*ppszOverFlowBuffer)
          delete [] *ppszOverFlowBuffer;
        
        //Remember the overflow buffer for the next time around
        *ppszOverFlowBuffer = pszNewBuf;        
      }
		}
	}

  return TRUE;
}
